import inspect

from django.core.exceptions import ObjectDoesNotExist

from djpb.registry import MODEL_TO_PROTO_CLS, PROTO_META, ProtoMeta
from djpb.serializers import SERIALIZERS, DEFAULT_SERIALIZER
from djpb.signals import pre_django_to_proto, post_django_to_proto
from djpb.stubs import DjModel, ProtoMsg
from djpb.util import (
    build_django_field_map,
    resolve_django_field_type,
    get_django_field_repr,
)


def django_to_proto_bytes(django_obj: DjModel, proto_obj: ProtoMsg = None) -> bytes:
    proto_obj = django_to_proto(django_obj, proto_obj)
    proto_bytes = proto_obj.SerializeToString()
    return proto_bytes


def django_to_proto(
    django_obj: DjModel, proto_obj: ProtoMsg = None, *, proto_meta: ProtoMeta = None
) -> ProtoMsg:
    django_model = type(django_obj)

    if proto_obj is None:
        try:
            proto_cls = MODEL_TO_PROTO_CLS[django_model][0]
        except IndexError:
            raise ValueError(
                f"Please specify at least one protobuf class for the model {django_model.__qualname__!r}."
            )
        proto_obj = proto_cls()

    field_map = build_django_field_map(django_obj)

    proto_meta = proto_meta or PROTO_META[django_model]
    custom = proto_meta.custom

    pre_django_to_proto.send(django_model, proto_obj=proto_obj, django_obj=django_obj)

    for proto_field in proto_obj.DESCRIPTOR.fields:
        field_name = proto_field.name

        # handle custom fields
        try:
            field = custom[field_name]
        except KeyError:
            pass
        else:
            field.update_proto(django_obj, proto_obj, field_name)
            continue

        # handle set_null
        if field_name.endswith("__set_null"):
            continue

        django_field_type = resolve_django_field_type(
            django_model, field_map, field_name
        )

        # check if proto field can be in an "unset" state
        try:
            proto_obj.HasField(field_name)
        except ValueError:
            can_unset = False
        else:
            can_unset = True

        try:
            value = getattr(django_obj, field_name)
        except ObjectDoesNotExist:
            if can_unset:
                # leave this field "unset"
                continue
            raise

        if value is None:
            if can_unset:
                # leave this field "unset"
                continue

            django_field_repr = get_django_field_repr(
                django_field_type, django_model, field_name
            )
            raise ValueError(
                f"Can't serialize None-type value for {django_field_repr}, "
                f"because protobuf doesn't support null types.\n"
                f"You can wrap the field with a `oneof` as a workaround."
            )

        # walk down the MRO to resolve the serializer for this field type
        serializer = DEFAULT_SERIALIZER
        for base_type in inspect.getmro(django_field_type):
            try:
                serializer = SERIALIZERS[base_type]
            except KeyError:
                continue
            else:
                break

        try:
            serializer.update_proto(proto_obj, field_name, value)
        except Exception as e:
            django_field_repr = get_django_field_repr(
                django_field_type, django_model, field_name
            )
            serializer_repr = repr(serializer.__class__.__qualname__)
            raise ValueError(
                f"Failed to serialize {django_field_repr} using {serializer_repr}."
            ) from e

    post_django_to_proto.send(django_model, proto_obj=proto_obj, django_obj=django_obj)

    return proto_obj



import msgpack
from rest_framework.renderers import BaseRenderer


class MsgpackRenderer(BaseRenderer):
    media_type = "application/msgpack"
    format = "msgpack"
    render_style = "binary"
    charset = None

    def render(self, data, media_type=None, renderer_context=None):
        return msgpack.packb(data, use_bin_type=True)



import inspect
import io
import typing
from contextlib import redirect_stdout
from textwrap import indent

from django.db import models
from django.db.models.fields.related_descriptors import (
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
    ManyToManyDescriptor,
    ForeignKeyDeferredAttribute,
)

from .registry import PROTO_META
from .stubs import DjField, DjFieldType, DjModelType
from .util import get_django_field_repr, build_django_field_map, disjoint

PROTO_TIMESTAMP_TYPE = "google.protobuf.Timestamp"
PROTO_STRUCT_TYPE = "google.protobuf.Struct"
PROTO_ANY_TYPE = "google.protobuf.Any"
PROTO_VALUE_TYPE = "google.protobuf.Value"

PROTO_IMPORTS = {
    PROTO_ANY_TYPE: "google/protobuf/any.proto",
    PROTO_VALUE_TYPE: "google/protobuf/struct.proto",
    PROTO_STRUCT_TYPE: "google/protobuf/struct.proto",
    PROTO_TIMESTAMP_TYPE: "google/protobuf/timestamp.proto",
}

DJANGO_TO_PROTO_FIELD_TYPE = {
    models.TextField: "string",
    models.CharField: "string",
    models.IntegerField: "int32",
    models.BooleanField: "bool",
    models.FloatField: "double",
}

RELATED_FIELD_TYPES = [
    models.ForeignKey,
    models.OneToOneField,
    models.ManyToManyField,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
    ManyToManyDescriptor,
]
RELATED_FIELD_TYPES_MANY = [
    models.ManyToManyField,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
]

SCALAR_FIELD_TYPES = {
    "double",
    "float",
    "int32",
    "int64",
    "uint32",
    "uint64",
    "sint32",
    "sint64",
    "fixed32",
    "fixed64",
    "sfixed32",
    "sfixed64",
    "bool",
    "string",
    "bytes",
}

ProtoFields = typing.Dict[str, typing.Tuple[DjField, str]]
ProtoModels = typing.Dict[DjFieldType, ProtoFields]


def gen_proto_for_models(dj_models: typing.Iterable[DjModelType]):
    proto_models: ProtoModels = {}
    imports = set()

    for model in dj_models:
        _gen_proto_for_model(model, proto_models)

    with io.StringIO() as f, redirect_stdout(f):
        for model, fields in proto_models.items():
            print("message %s {" % model.__name__)

            for i, (field_name, (field, proto_type)) in enumerate(fields.items()):
                line = f"{proto_type} {field_name} = {i + 1};"

                oneof = (
                    getattr(field, "null", False) and proto_type in SCALAR_FIELD_TYPES
                )
                if oneof:
                    line = indent(line, " " * 4)
                    line = "oneof __%s_oneof {\n%s\n}" % (field_name, line)

                line = indent(line, " " * 4)
                print(line)

                try:
                    imports.add(PROTO_IMPORTS[proto_type])
                except KeyError:
                    pass

            print("}\n")

        body = f.getvalue()

    with io.StringIO() as f, redirect_stdout(f):
        print('syntax = "proto3";\n')

        for i in imports:
            print(f'import "{i}";')
        print()

        body = f.getvalue() + body

    body = body.strip()
    return body


def _gen_proto_for_model(model: DjModelType, proto_models: ProtoModels):
    if model in proto_models:
        return
    proto_models[model] = {}

    proto_meta = PROTO_META[model]

    field_map = build_django_field_map(model)

    if proto_meta.fields:
        assert not (
            proto_meta.exclude or proto_meta.extra
        ), "'exclude' and 'extra' are not allowed if 'fields' is specified.."

        field_map = {
            name: field_map.get(name) or getattr(model, name)
            for name in proto_meta.fields
        }
    else:
        assert disjoint(
            proto_meta.exclude, proto_meta.extra
        ), "'exclude' and 'extra' must be disjoint sets."

        # exclude the primary key by default
        pk_name = model._meta.pk.name
        if pk_name not in proto_meta.extra:
            del field_map[pk_name]

        for name in proto_meta.exclude:
            del field_map[name]

        for name in proto_meta.extra:
            if name in field_map:
                continue
            field_map[name] = getattr(model, name)

    proto_fields = {
        name: _resolve_proto_type(name, field, model, proto_models)
        for name, field in field_map.items()
    }

    # add custom fields
    custom = getattr(proto_meta, "custom", {})
    proto_fields.update(
        {name: (field, field.proto_type) for name, field in custom.items()}
    )

    proto_models[model] = proto_fields


def _resolve_proto_type(
    field_name: str, field, model: DjModelType, proto_models: ProtoModels
) -> typing.Tuple[DjField, str]:
    field_type = type(field)

    proto_meta = PROTO_META[model]

    if field_name in proto_meta.enums:
        enum_cls = proto_meta.enums[field_name]
        enum_clsname = enum_cls.DESCRIPTOR.name
        return field, enum_clsname

    if issubclass(field_type, ForeignKeyDeferredAttribute):
        field = field.field
        model = field.related_model
        pk_field = model._meta.pk
        proto_type = _proto_type_for_field(pk_field.name, type(pk_field), model)

    elif field_type in RELATED_FIELD_TYPES:
        try:
            related_model = field.related_model
        except AttributeError:
            related_model = field.field.model

        _gen_proto_for_model(related_model, proto_models)

        proto_type = related_model.__name__
        if field_type in RELATED_FIELD_TYPES_MANY:
            proto_type = f"repeated {proto_type}"

    else:
        proto_type = _proto_type_for_field(field_name, field_type, model)

    return field, proto_type


def _proto_type_for_field(
    field_name: str, field_type: DjFieldType, model: DjModelType,
) -> str:
    # walk down the MRO to resolve the field type
    proto_type = None
    for base_type in inspect.getmro(field_type):
        try:
            proto_type = DJANGO_TO_PROTO_FIELD_TYPE[base_type]
        except KeyError:
            continue
        else:
            break
    if proto_type is None:
        django_field_repr = get_django_field_repr(field_type, model, field_name)
        raise ValueError(
            f"Could not find a suitable protobuf type for {django_field_repr}."
        )
    return proto_type



import typing

from django.db.models.fields.related_descriptors import ForeignKeyDeferredAttribute

from djpb.stubs import DjModel, DjField, DjModelType, DjFieldType, ProtoMsg

DjangoFieldMap = typing.Dict[str, DjField]


def build_django_field_map(django_obj: DjModel) -> DjangoFieldMap:
    options = django_obj._meta
    fields = options.fields + options.many_to_many
    field_map = {f.name: f for f in fields}
    return field_map


def resolve_django_field_type(
    django_model: DjModelType, field_map: DjangoFieldMap, field_name: str
) -> DjFieldType:
    django_field = None
    try:
        django_field = field_map[field_name]
    except KeyError:
        try:
            django_field = getattr(django_model, field_name)
        except AttributeError:
            pass

    if django_field is None:
        raise ValueError(
            f"Protobuf field {field_name!r} does not exist "
            f"in Django model {django_model.__qualname__!r}."
        )

    if isinstance(django_field, ForeignKeyDeferredAttribute):
        related_model = django_field.field.related_model
        django_field = related_model._meta.pk

    django_field_type = type(django_field)
    return django_field_type


def get_django_field_repr(
    django_field_type: DjFieldType, django_model: DjModelType, field_name: str,
) -> str:
    return f"field '{django_model.__qualname__}.{field_name}' of type {django_field_type.__qualname__!r}"


def disjoint(x, y):
    return set(x).isdisjoint(y)


def create_proto_field_obj(proto_obj: ProtoMsg, field_name: str) -> ProtoMsg:
    fields = {field.name: field for field in proto_obj.DESCRIPTOR.fields}
    field = fields[field_name]
    return field.message_type._concrete_class()



from django.core.exceptions import ValidationError
from rest_framework import serializers
from rest_framework.fields import get_error_detail

from ..django_to_proto import django_to_proto
from ..proto_to_django import proto_to_django
from ..registry import MODEL_TO_PROTO_CLS
from ..stubs import DjModelType, ProtoMsgType, DjModel, ProtoMsg


class ProtobufSerializer(serializers.BaseSerializer):
    model: DjModelType
    proto_cls: ProtoMsgType = None
    do_full_clean: bool = True

    @staticmethod
    def for_model(
        model: DjModelType,
        *,
        proto_cls: ProtoMsgType = None,
        do_full_clean: bool = True,
    ):
        class ProtobufSerializerForModel(ProtobufSerializer):
            pass

        ProtobufSerializerForModel.model = model
        ProtobufSerializerForModel.proto_cls = proto_cls
        ProtobufSerializerForModel.do_full_clean = do_full_clean

        return ProtobufSerializerForModel

    def to_internal_value(self, data):
        return data

    def create(self, validated_data: bytes) -> DjModel:
        return self.update(self.model(), validated_data)

    def update(self, instance: DjModel, validated_data: bytes) -> DjModel:
        proto_cls = self.get_proto_cls()
        proto_obj = proto_cls.FromString(validated_data)
        return self._from_proto(proto_obj, instance)

    def _from_proto(self, proto_obj: ProtoMsg, instance: DjModel) -> DjModel:
        try:
            return proto_to_django(
                proto_obj, instance, do_full_clean=self.do_full_clean,
            )
        except ValidationError as e:
            raise serializers.ValidationError(get_error_detail(e))

    def to_representation(self, instance: DjModel) -> bytes:
        proto_obj = self._to_proto(instance)
        return proto_obj.SerializeToString()

    def _to_proto(self, instance: DjModel) -> ProtoMsg:
        proto_cls = self.get_proto_cls()
        proto_obj = proto_cls()
        proto_obj = django_to_proto(instance, proto_obj)
        return proto_obj

    def get_proto_cls(self) -> ProtoMsgType:
        if self.proto_cls:
            return self.proto_cls
        try:
            return MODEL_TO_PROTO_CLS[self.model][0]
        except IndexError:
            raise ValueError(
                f"Please specify at least one protobuf class for the model {self.model.__qualname__!r}."
            )

    #
    # Mostly copy-pasted from the parent class,
    # except a few changes to allow protobuf types to work nicely
    #
    def save(self, **kwargs):
        assert not hasattr(self, "save_object"), (
            "Serializer `%s.%s` has old-style version 2 `.save_object()` "
            "that is no longer compatible with REST framework 3. "
            "Use the new-style `.create()` and `.update()` methods instead."
            % (self.__class__.__module__, self.__class__.__name__)
        )

        assert hasattr(
            self, "_errors"
        ), "You must call `.is_valid()` before calling `.save()`."

        assert (
            not self.errors
        ), "You cannot call `.save()` on a serializer with invalid data."

        # Guard against incorrect use of `serializer.save(commit=False)`
        assert "commit" not in kwargs, (
            "'commit' is not a valid keyword argument to the 'save()' method. "
            "If you need to access data before committing to the database then "
            "inspect 'serializer.validated_data' instead. "
            "You can also pass additional keyword arguments to 'save()' if you "
            "need to set extra attributes on the saved model instance. "
            "For example: 'serializer.save(owner=request.user)'.'"
        )

        assert not hasattr(self, "_data"), (
            "You cannot call `.save()` after accessing `serializer.data`."
            "If you need to access data before committing to the database then "
            "inspect 'serializer.validated_data' instead. "
        )

        if self.instance is not None:
            self.instance = self.update(self.instance, self.validated_data)
            assert (
                self.instance is not None
            ), "`update()` did not return an object instance."
        else:
            self.instance = self.create(self.validated_data)
            assert (
                self.instance is not None
            ), "`create()` did not return an object instance."

        return self.instance



from .custom_field import (
    CustomField,
    ReadOnlyValueField,
    ReadOnlyQuerySetField,
    ReadOnlyQueryStrField,
)
from .django_to_proto import django_to_proto, django_to_proto_bytes
from .gen_proto import gen_proto_for_models
from .proto_to_django import proto_to_django
from .registry import register_model, ProtoMeta



from django.db import transaction

from djpb.django_to_proto import SERIALIZERS, DEFAULT_SERIALIZER
from djpb.registry import PROTO_CLS_TO_MODEL, MODEL_TO_PROTO_CLS, PROTO_META
from djpb.serializers import FieldSerializer, SaveNode
from djpb.util import (
    build_django_field_map,
    resolve_django_field_type,
    get_django_field_repr,
)
from .signals import post_proto_to_django, pre_proto_to_django
from .stubs import ProtoMsg, DjModel, DjModelType


def proto_to_django(
    proto_obj: ProtoMsg, django_obj: DjModel = None, *, do_full_clean=False
) -> DjModel:
    node = _proto_to_django(proto_obj, django_obj)
    with transaction.atomic():
        node.save(do_full_clean)
    django_obj = node.django_obj
    return django_obj


def _proto_to_django(proto_obj: ProtoMsg, django_obj: DjModel = None) -> SaveNode:
    proto_fields = {x.name: x for x in proto_obj.DESCRIPTOR.fields}

    if django_obj is None:
        proto_cls = type(proto_obj)
        django_cls = PROTO_CLS_TO_MODEL[proto_cls]

        # try to get an existing object if it's pk is present in the proto fields
        pk_field_name = django_cls._meta.pk.name

        if pk_field_name in proto_fields:
            pk = getattr(proto_obj, pk_field_name)

            if pk:  # pk might be 0, let's ignore that
                try:
                    django_obj = django_cls.objects.get(**{pk_field_name: pk})
                except django_cls.DoesNotExist:
                    pass

        if django_obj is None:
            django_obj = django_cls()

    django_model = django_obj.__class__
    field_map = build_django_field_map(django_obj)
    node = SaveNode(django_obj)

    proto_meta = PROTO_META[django_model]
    custom = proto_meta.custom

    pre_proto_to_django.send(django_model, proto_obj=proto_obj, django_obj=django_obj)

    for field_name, proto_field in proto_fields.items():
        # determine if field is in "unset" state
        try:
            unset = not proto_obj.HasField(field_name)
        except ValueError:
            unset = False

        if unset:
            # leave "unset" fields as-is
            continue

        # handle custom fields
        try:
            field = custom[field_name]
        except KeyError:
            pass
        else:
            field.update_django(node, proto_obj, field_name)
            continue

        # handle set_null
        if field_name.endswith("__set_null"):
            field_name = field_name[: -len("__set_null")]
            setattr(django_obj, field_name, None)
            continue

        django_field_type = resolve_django_field_type(
            django_model, field_map, field_name
        )
        value = getattr(proto_obj, field_name)

        serializer: FieldSerializer = SERIALIZERS.get(
            django_field_type, DEFAULT_SERIALIZER
        )
        try:
            serializer.update_django(node, field_name, value)
        except Exception as e:
            django_field_repr = get_django_field_repr(
                django_field_type, django_model, field_name
            )
            serializer_repr = repr(serializer.__class__.__qualname__)
            raise ValueError(
                f"Failed to de-serialize {django_field_repr} using {serializer_repr}."
            ) from e

    post_proto_to_django.send(django_model, proto_obj=proto_obj, django_obj=django_obj)

    return node



import typing

from django.db import models
from google.protobuf.message import Message

DjModel = typing.TypeVar("DjModel", bound=models.Model)
DjModelType = typing.TypeVar("DjModelType", bound=typing.Type[models.Model])

DjField = typing.TypeVar("DjField", bound=models.Field)
DjFieldType = typing.TypeVar("DjFieldType", bound=typing.Type[models.Field])

ProtoMsg = typing.TypeVar("ProtoMsg", bound=Message)
ProtoMsgType = typing.TypeVar("ProtoMsgType", bound=typing.Type[Message])



import typing as T
import uuid
from dataclasses import dataclass

from django.conf import settings
from django.db import models
from django.db.models.fields.related_descriptors import ReverseManyToOneDescriptor
from django.utils import timezone
from google.protobuf.json_format import MessageToDict, ParseDict
from google.protobuf.struct_pb2 import Value

from djpb.util import get_django_field_repr, create_proto_field_obj
from .gen_proto import (
    DJANGO_TO_PROTO_FIELD_TYPE,
    PROTO_VALUE_TYPE,
    PROTO_TIMESTAMP_TYPE,
)
from .stubs import DjModel, ProtoMsg, DjFieldType


def register_serializer(cls: T.Type["FieldSerializer"]) -> T.Type["FieldSerializer"]:
    assert issubclass(cls, FieldSerializer)

    proto_type_name = getattr(cls, "proto_type_name", None)

    for field_type in cls.field_types:
        SERIALIZERS[field_type] = cls()
        if proto_type_name:
            DJANGO_TO_PROTO_FIELD_TYPE[field_type] = proto_type_name

    return cls


class FieldSerializer:
    """
    The default serializer that does no serialization, and directly calls setattr.

    All serializers must subclass FieldSerializer,
    and register themselves using the @register_serializer decorator.
    """

    field_types: T.Iterable[DjFieldType]
    proto_type_name: str

    def update_proto(self, proto_obj: ProtoMsg, field_name: str, value):
        setattr(proto_obj, field_name, value)

    def update_django(self, node: "SaveNode", field_name: str, value):
        setattr(node.django_obj, field_name, value)


DEFAULT_SERIALIZER = FieldSerializer()
SERIALIZERS: T.Dict[DjFieldType, FieldSerializer] = {}


@register_serializer
class DateTimeFieldSerializer(FieldSerializer):
    field_types = (models.DateTimeField,)
    proto_type_name = PROTO_TIMESTAMP_TYPE

    def update_proto(self, proto_obj, field_name, value):
        field = getattr(proto_obj, field_name)
        field.FromDatetime(value)

    def update_django(self, node, field_name, value):
        value = value.ToDatetime()
        value = value.replace(tzinfo=timezone.utc)
        super().update_django(node, field_name, value)


@register_serializer
class UUIDFieldSerializer(FieldSerializer):
    field_types = (models.UUIDField,)
    proto_type_name = "string"

    def update_proto(self, proto_obj, field_name, value):
        value = str(value)
        super().update_proto(proto_obj, field_name, value)

    def update_django(self, node, field_name, value):
        value = uuid.UUID(value)
        super().update_django(node, field_name, value)


@register_serializer
class FileFieldSerializer(FieldSerializer):
    field_types = (models.FileField,)
    proto_type_name = "string"

    @property
    def use_url(self):
        try:
            return settings.DJPB_FILE_FIELD_USE_URL
        except AttributeError:
            return False

    def update_proto(self, proto_obj, field_name, value):
        if self.use_url:
            try:
                value = value.url
            except ValueError:
                value = ""
        else:
            value = str(value)

        super().update_proto(proto_obj, field_name, value)

    def update_django(self, node, field_name, value):
        if "://" in value:
            django_field_repr = get_django_field_repr(
                models.FileField, node.django_obj.__class__, field_name
            )
            raise ValueError(
                f"Please provide the file path, not the URL for {django_field_repr} = {value!r}."
            )

        super().update_django(node, field_name, value)


@register_serializer
class JSONFieldSerializer(FieldSerializer):
    field_types = (models.JSONField,)
    proto_type_name = PROTO_VALUE_TYPE

    def update_proto(self, proto_obj, field_name, value):
        value = ParseDict(value, Value())
        field = getattr(proto_obj, field_name)
        field.CopyFrom(value)

    def update_django(self, node, field_name, value):
        value = MessageToDict(value)
        super().update_django(node, field_name, value)


class DeferredSerializer(FieldSerializer):
    is_many = False

    def update_django(self, node, field_name, value):
        from djpb.proto_to_django import _proto_to_django

        if self.is_many:
            rel_pb_objs = list(value)
        else:
            rel_pb_objs = [value]

        if self.is_many and node.django_obj.id:
            rel_objs = getattr(node.django_obj, field_name).all()
        else:
            rel_objs = None

        to_keep = []

        for pb_obj in rel_pb_objs:
            if rel_objs and hasattr(pb_obj, "id") and pb_obj.id:
                dj_obj = rel_objs.get(id=pb_obj.id)
                to_keep.append(pb_obj.id)
            else:
                dj_obj = None
            child_node = _proto_to_django(pb_obj, dj_obj)
            node.add_child(SaveNodeChild(self, field_name, child_node))

        if rel_objs:
            rel_objs.exclude(id__in=to_keep).delete()

    def save(
        self,
        django_obj: DjModel,
        field_name: str,
        child_node: "SaveNode",
        do_full_clean: bool,
    ):
        ...


@register_serializer
class OneToXSerializer(DeferredSerializer):
    field_types = (models.OneToOneField, models.ForeignKey)

    def update_proto(self, proto_obj, field_name, value):
        from djpb import django_to_proto

        field = getattr(proto_obj, field_name)
        value = django_to_proto(value, create_proto_field_obj(proto_obj, field_name))
        field.CopyFrom(value)

    def save(self, django_obj, field_name, child_node, do_full_clean):
        # save child object
        child_node.save(do_full_clean)
        rel_obj = child_node.django_obj

        # set parent object's field to child object
        setattr(django_obj, field_name, rel_obj)

        # save parent object
        if do_full_clean:
            django_obj.full_clean()
        django_obj.save()


class ManyToXSerializer(DeferredSerializer):
    is_many = True

    def update_proto(self, proto_obj, field_name, value):
        from djpb import django_to_proto

        field = getattr(proto_obj, field_name)
        msgs = [
            django_to_proto(obj, create_proto_field_obj(proto_obj, field_name))
            for obj in value.all()
        ]
        del field[:]
        field.extend(msgs)


@register_serializer
class ManyToOneSerializer(ManyToXSerializer):
    field_types = (ReverseManyToOneDescriptor,)

    def save(self, django_obj, field_name, child_node, do_full_clean):
        # save parent object
        if do_full_clean:
            django_obj.full_clean()
        django_obj.save()

        # get child object
        rel_obj = child_node.django_obj

        # set child object's field to parent object
        rel_manager = getattr(django_obj, field_name)
        rel_name = rel_manager.field.name
        setattr(rel_obj, rel_name, django_obj)

        # save related object
        child_node.save(do_full_clean)


@register_serializer
class ManyToManySerializer(ManyToXSerializer):
    field_types = (models.ManyToManyField,)

    def save(self, django_obj, field_name, child_node, do_full_clean):
        # save child object
        child_node.save(do_full_clean)

        # save parent object
        if do_full_clean:
            django_obj.full_clean()
        django_obj.save()

        # add child to parent's m2m manager
        rel_manager = getattr(django_obj, field_name)
        rel_obj = child_node.django_obj
        rel_manager.add(rel_obj)


class SaveNodeChild(T.NamedTuple):
    serializer: DeferredSerializer
    field_name: str
    node: "SaveNode"

    def __hash__(self) -> int:
        return hash(self.node)


@dataclass
class SaveNode:
    django_obj: DjModel

    def __post_init__(self):
        self._children: T.Set[SaveNodeChild] = set()

    def __hash__(self) -> int:
        return id(self.django_obj)

    def add_child(self, child: SaveNodeChild):
        self._children.add(child)

    def save(self, do_full_clean: bool):
        if self._children:
            for serializer, field_name, child_node in self._children:
                serializer.save(self.django_obj, field_name, child_node, do_full_clean)
        else:
            if do_full_clean:
                self.django_obj.full_clean()
            self.django_obj.save()



import msgpack
from msgpack import UnpackException
from rest_framework.exceptions import ParseError
from rest_framework.parsers import BaseParser


class MsgpackParser(BaseParser):
    media_type = "application/msgpack"

    def parse(self, stream, media_type=None, parser_context=None):
        try:
            return msgpack.load(stream)
        except UnpackException as e:
            raise ParseError(f"MessagePack parse error - {e!r}") from e



import typing
from collections import defaultdict

from .stubs import DjModelType, ProtoMsgType

# to avoid circular import
if False:
    from .custom_field import CustomField

MODEL_TO_PROTO_CLS: typing.DefaultDict[
    DjModelType, typing.List[ProtoMsgType]
] = defaultdict(lambda: [])

PROTO_CLS_TO_MODEL: typing.Dict[ProtoMsgType, DjModelType] = {}


class ProtoMeta:
    def __init__(
        self,
        exclude: typing.Sequence[str] = (),
        extra: typing.Sequence[str] = (),
        fields: typing.Sequence[str] = (),
        custom: typing.Dict[str, "CustomField"] = None,
        enums: typing.Dict[str, typing.Type] = None,
    ):
        if custom is None:
            custom = {}
        if enums is None:
            enums = {}
        self.exclude = exclude
        self.extra = extra
        self.fields = fields
        self.custom = custom
        self.enums = enums


PROTO_META: typing.DefaultDict[DjModelType, ProtoMeta] = defaultdict(ProtoMeta)


def register_model(
    proto_classes: typing.List[ProtoMsgType], proto_meta: ProtoMeta = None
):
    def decorator(django_model: DjModelType):
        MODEL_TO_PROTO_CLS[django_model] = proto_classes
        if proto_meta:
            PROTO_META[django_model] = proto_meta
        for proto_class in proto_classes:
            PROTO_CLS_TO_MODEL[proto_class] = django_model

        return django_model

    return decorator



from django.core.management.base import BaseCommand

from djpb.gen_proto import gen_proto_for_models
from djpb.registry import MODEL_TO_PROTO_CLS


class Command(BaseCommand):
    help = "Generate protobuf file for all registered models."

    def handle(self, *args, **options):
        print(gen_proto_for_models(MODEL_TO_PROTO_CLS.keys()))



import typing
from dataclasses import dataclass

from .django_to_proto import django_to_proto
from .serializers import SaveNode
from .stubs import DjModel, ProtoMsg
from .util import create_proto_field_obj


@dataclass
class CustomField:
    proto_type: str
    null: bool = False

    def update_proto(self, django_obj: DjModel, proto_obj: ProtoMsg, field_name: str):
        pass

    def update_django(self, node: "SaveNode", proto_obj: ProtoMsg, field_name: str):
        pass


@dataclass
class ReadOnlyQueryStrField(CustomField):
    query: str = None

    def update_proto(self, django_obj, proto_obj, field_name):
        model = type(django_obj)
        value = (
            model.objects.filter(pk=django_obj.pk)
            .values_list(self.query or field_name, flat=True)
            .first()
        )
        if value is None:
            return
        setattr(proto_obj, field_name, value)


@dataclass
class ReadOnlyQuerySetField(CustomField):
    get_queryset: typing.Callable = None

    def update_proto(self, django_obj, proto_obj, field_name):
        field = getattr(proto_obj, field_name)
        msgs = [
            django_to_proto(obj, create_proto_field_obj(proto_obj, field_name))
            for obj in self.get_queryset(django_obj)
        ]
        field.extend(msgs)


@dataclass
class ReadOnlyValueField(CustomField):
    get_value: typing.Callable = None

    def update_proto(self, django_obj, proto_obj, field_name):
        value = self.get_value(django_obj)
        setattr(proto_obj, field_name, value)
